"ref_diff.py -- run sql diffs on git repos"

import sqlparse, collections, os, pathlib
from . import githelp, diffing

def files_to_smts(fulltexts):
  "given a list of file contents, read them, parse them and return list of statements"
  stmts = []
  for text in fulltexts:
    stmts.extend(sqlparse.parse(text))
  return stmts

def ref_diff(repo, ref1, ref2, pattern):
  "given repo, refs & glob pattern, return list of uniquely identified statements making up diff (or uniquely ID'd errors)"
  contents = (
    githelp.get_streams(repo.commit(ref).tree, pattern)
    for ref in (ref1, ref2)
  )
  left, right = [
    files_to_smts(fulltexts)
    for fulltexts in contents
  ]
  return diffing.diff(left, right)

def ref_range_diff(repo, ref1, ref2, pattern, opaque=False):
  "run ref_diff() once per intermediate commit for commits who change files matching pattern"
  assert not os.path.isabs(pattern), "we don't know how to transform glob to absolute path -- file a bug"
  assert os.path.isabs(repo.working_dir), "working dir is non-abs -- file a bug"
  # note: pathlib '/' operator should filter out single-dot
  absglob = pathlib.Path.cwd() / pathlib.Path(pattern)
  assert not any(part == '.' or part == '..' for part in pathlib.Path(absglob).parts), "parent paths not supported -- file a bug"
  commits = list(repo.iter_commits(
    f'{ref1}...{ref2}',
    paths=str(absglob)
  ))
  if opaque:
    # note: this is intended to be len=0 when change list is empty
    commits = [repo.commit(ref1)] + commits[:1]
  else:
    commits.append(repo.commit(ref1))
    commits = list(reversed(commits))
  return collections.OrderedDict([
    [right.hexsha, ref_diff(repo, left.hexsha, right.hexsha, str(absglob))]
    for left, right in zip(commits[:-1], commits[1:])
  ])

def extract_errors(sha_table_stmts):
  ret = {
    sha: diffing.get_errors(table_stmts)
    for sha, table_stmts in sha_table_stmts.items()
  }
  for k, v in list(ret.items()):
    if not v:
      del ret[k]
  return ret

def try_repair_errors(errors, manual_overrides, sha_table_stmts):
  """Try to repair this.
  errors is {sha: {table: error_list}}
  manual_overrides is {sha: {table: stmts}}
  sha_table_stmts is {sha: {table: stmts}}
  returns remaining_errors, mutates sha_table_stmts.
  """
  for sha, table_errors in errors.items():
    for table in table_errors:
      if sha in manual_overrides and table in manual_overrides[sha]:
        sha_table_stmts[sha][table] = manual_overrides[sha][table]
  return extract_errors(sha_table_stmts)



"diffing.py -- sql-diffing"

import collections, sqlparse, yaml
from . import wrappers

class DiffError(Exception): pass

def group_by_table(stmts):
  "take list of WrappedStatement"
  groups = collections.defaultdict(list)
  for stmt in stmts:
    if isinstance(stmt, (wrappers.CreateTable, wrappers.CreateIndex)):
      groups[stmt.table].append(stmt)
    elif stmt is None:
      pass # where are these coming from? it happens in the test suite even
    else:
      raise DiffError("unhandled type", type(stmt))
  return groups

class UnsupportedChange(Exception):
  "returned in place of a migration string when there's an error"

def diff_column(table, colname, left, right):
  "return list of stmts to alter column, or UnsupportedChange if we're confused"
  if not left.success or not right.success:
    # todo: add details (column name and rendered old/new)
    return [UnsupportedChange(f"the column parser failed on new or old change for col {colname}")]
  assert left.name == right.name
  ret = []
  prefix = f"alter table {table} alter column {left.name}"
  if left.type != right.type:
    ret.append(f"{prefix} type {right.type};")
  if left.default != right.default:
    if right.default is None:
      ret.append(UnsupportedChange("can't unset default, file a bug and workaround by setting `default null` for now"))
    else:
      ret.append(f"{prefix} set default {right.default};")
  if left.unique != right.unique:
    ret.append(UnsupportedChange("can't modify uniqueness, file a bug"))
  if left.not_null != right.not_null:
    # note: I think the possible values here are (None | True), shouldn't ever be False
    # todo: refactor this to `null` and include null / not_null as sources
    # todo: this is assuming that 'not specified' is nullable -- link to DB docs supporting this and figure out the pkey case
    ret.append(f"{prefix} {'set' if right.not_null else 'drop'} not null;")
  return ret

def diff_stmt(left, right):
  "diff two WrappedStmt with same unique key. return list of statements to run."
  assert left.unique == right.unique
  table = left.table
  if isinstance(left, wrappers.CreateTable):
    left_cols = {col.name: col for col in left.columns()}
    right_cols = {col.name: col for col in right.columns()}
    changes = [
      f'alter table {table} add column {right_cols[k].render()};'
      for k in right_cols if k not in left_cols
    ]
    changed = {
      k: (left_cols[k], right_cols[k])
      for k in right_cols
      if k in left_cols and left_cols[k] != right_cols[k]
    }
    if changed:
      for left_col, right_col in changed.values():
        changes.extend(diff_column(table, left_col.name, left_col.parse(), right_col.parse()))
    for k in left_cols:
      if k not in right_cols:
        changes.append(f'alter table {table} drop column {k};')
    if left.tail() != right.tail():
      change = ' '.join([expr.value for expr in left.tail() or right.tail()])
      changes.append(UnsupportedChange(f"can't modify table suffix: `{change}`"))
    if left.pkey_fields() != right.pkey_fields():
      # note: order matters here too; don't compare sets
      if left.pkey_fields():
        changes.append(f'alter table {table} drop constraint {table}_pkey;')
      new_pkey = ', '.join(right.pkey_fields())
      if new_pkey:
        changes.append(f'alter table {table} add primary key ({new_pkey});')
    return changes
  else:
    raise DiffError("unhandled type", type(left))

def diff_stmts(left, right):
  "takes WrappedStmt lists, all for same table, and compares them"
  key_l = {stmt.unique: stmt for stmt in left}
  key_r = {stmt.unique: stmt for stmt in right}
  output = []
  for k in key_r:
    if k not in key_l:
      output.append(str(key_r[k].stmt).strip())
    elif key_l[k] == key_r[k]:
      pass # not relevant to diff
    else:
      output.extend(diff_stmt(key_l[k], key_r[k]))
  return output

def diff(left, right):
  """take two lists of statements.
  return dict of {tablename: list of migration statements or errors}
  """
  # todo: figure out some way to apply statements in order even across tables
  # realistically this needs to be a database, not a nested dict -- it gets queried in a lot of ways
  groups_l = group_by_table(map(wrappers.wrap, left))
  groups_r = group_by_table(map(wrappers.wrap, right))
  output = collections.OrderedDict()
  for key, stmts in groups_r.items():
    if key in groups_l:
      changes = [stmt for stmt in diff_stmts(groups_l[key], stmts)]
      if changes:
        output[key] = changes
    else:
      output[key] = [str(wrapped.stmt).strip() for wrapped in stmts]
  return output

def get_errors(table_stmt_dict):
  return {
    table: [stmt for stmt in stmts if isinstance(stmt, Exception)]
    for table, stmts in table_stmt_dict.items()
    if any(isinstance(stmt, Exception) for stmt in stmts)
  }



import pytest, sqlparse, collections
from automig.lib import diffing, wrappers

CREATE_TABLE = [
  'create table t1 (a int);',
  'create table t1 (a int); create table t2 (a int);',
]

def test_create_table():
  delta = diffing.diff(*map(sqlparse.parse, CREATE_TABLE))
  assert delta == {'t2': ['create table t2 (a int);']}

@pytest.mark.skip
def test_drop_table():
  "tests that removed tables *aren't* dropped -- there's like no reason to right?"
  raise NotImplementedError

ADD_COLUMN = [
  'create table t1 (a int primary key);',
  'create table t1 (a int primary key, b int);',
]

def test_add_column():
  delta = diffing.diff(*map(sqlparse.parse, ADD_COLUMN))
  assert delta == {'t1': ['alter table t1 add column b int;']}

MODIFY_COLUMN = [
  'create table t1 (a jsonb primary key, b int default 10, c text[], d varchar(12));',
  # this is modifying a default, setting not nullable, and changing a varchar size
  # the array column is to test Column.parse() on array types
  # the jsonb column is because jsonb is a keyword not a builtin, broke the parser initially
  'create table t1 (a jsonb primary key, b int default 20, c text[] not null, d varchar(24));',
]

def test_modify_column():
  assert diffing.diff(*map(sqlparse.parse, MODIFY_COLUMN)) == {'t1': [
    'alter table t1 alter column b set default 20;',
    'alter table t1 alter column c set not null;',
    'alter table t1 alter column d type varchar(24);',
  ]}

@pytest.mark.skip
def test_diff_column():
  "directly test diff_column cases"
  raise NotImplementedError

def parse_column(type_):
  stmt, = sqlparse.parse(f"create table t1 (x {type_})")
  col, = wrappers.wrap(stmt).columns()
  return col.parse()

def test_column_parser():
  "directly test wrappers.Column.parse() cases"
  for type_ in ['int', 'jsonb', 'text', 'varchar(12)', 'int[]', 'text[]']:
    assert parse_column(type_) == wrappers.ParsedColumn(True, 'x', type_)
    assert parse_column(type_ + ' not null') == wrappers.ParsedColumn(True, 'x', type_, not_null=True)
    assert parse_column(type_ + ' unique') == wrappers.ParsedColumn(True, 'x', type_, unique=True)
    assert parse_column(type_ + ' default now()') == wrappers.ParsedColumn(True, 'x', type_, default='now()')
    assert parse_column(type_ + ' default 20') == wrappers.ParsedColumn(True, 'x', type_, default='20')
    assert parse_column(type_ + " default 'hello'") == wrappers.ParsedColumn(True, 'x', type_, default="'hello'")
    assert parse_column(type_ + ' not null unique default 20') == wrappers.ParsedColumn(True, 'x', type_, not_null=True, unique=True, default='20')
  # make sure it doesn't treat 'primary key' as a column
  assert [col.name for col in wrappers.wrap(sqlparse.parse("create table t1 (a int, b int, primary key (a,b))")[0]).columns()] == ['a', 'b']

def test_pkey_fields():
  assert ['a'] == wrappers.wrap(sqlparse.parse("create table t1 (a int primary key, b int)")[0]).pkey_fields()
  assert ['a', 'b'] == wrappers.wrap(sqlparse.parse("create table t1 (a int, b int, primary key (a,b))")[0]).pkey_fields()

DROP_COLUMN = [
  'create table t1 (a int primary key, b int, c int);',
  'create table t1 (a int primary key, c int);',
]

def test_drop_column():
  delta = diffing.diff(*map(sqlparse.parse, DROP_COLUMN))
  assert delta == {'t1': ['alter table t1 drop column b;']}

MODIFY_KEY_1 = [
  'create table t1 (a int primary key);',
  'create table t1 (a int, primary key (a));',
]

MODIFY_KEY_2 = [
  'create table t1 (a int primary key);',
  'create table t1 (a int, b int, primary key (a, b));',
]

MODIFY_KEY_3 = [
  'create table t1 (a int primary key);',
  'create table t1 (a int);',
]

def test_modify_key():
  assert not diffing.diff(*map(sqlparse.parse, MODIFY_KEY_1))
  assert diffing.diff(*map(sqlparse.parse, MODIFY_KEY_2))['t1'] == [
    'alter table t1 add column b int;',
    'alter table t1 drop constraint t1_pkey;',
    'alter table t1 add primary key (a, b);',
  ]
  assert diffing.diff(*map(sqlparse.parse, MODIFY_KEY_3))['t1'] == ['alter table t1 drop constraint t1_pkey;']
  assert diffing.diff(*map(sqlparse.parse, reversed(MODIFY_KEY_3)))['t1'] == ['alter table t1 add primary key (a);',]

CREATE_INDEX = [
  'create unique index idx_col1 on t1 (col1);',
  'create unique index idx_col1 on t1 (col1); create unique index idx_col2 on t1 (col2);',
]

def test_add_index():
  delta = diffing.diff(*map(sqlparse.parse, CREATE_INDEX))
  assert delta == {'t1': ['create unique index idx_col2 on t1 (col2);']}

@pytest.mark.skip
def test_all_caps_keywords():
  raise NotImplementedError

NEWLINE = [
  'create table whatever (\n  a int\n);',
  'create table whatever (\n  a int,\n  b int\n);',
]

def test_newline():
  # this isn't asserting anything -- checking for a bug which caused a crash
  diffing.diff(*map(sqlparse.parse, NEWLINE))

COMMENT = [
  'create table whatever (\n  a int -- hello\n);',
  'create table whatever (\n  a int, -- hello\n  b int\n);',
]

def test_comments():
  # not asserting, just checking for crash
  diffing.diff(*map(sqlparse.parse, COMMENT))

PARTITION = [
  'create table whatever (a boolean primary key);',
  'create table whatever (a boolean primary key) partition by range (a);',
]

def test_tail():
  "directly test CreateTable.tail()"
  no, yes = [wrappers.wrap(parsed[0]) for parsed in map(sqlparse.parse, PARTITION)]
  assert no.tail() == []
  assert ['partition', 'by', 'range (a)'] == [tok.value for tok in yes.tail()]

def test_partition():
  assert diffing.diff(*map(sqlparse.parse, PARTITION))['whatever'][0].args == \
    ("can't modify table suffix: `partition by range (a)`",)



#!/usr/bin/env python3
"sa_harness.py -- create sqlalchemy definitions from create table & index stmts"

import sqlparse, glob, collections, argparse
from . import wrappers, diffing

PREAMBLE = """# autogenerated by sa_harness.py
import sqlalchemy as sa"""

def read_glob_stmts(glob_pattern):
  "return wrapped stmts for files matching glob"
  stmts = []
  for fname in glob.glob(glob_pattern):
    stmts.extend(map(wrappers.wrap, sqlparse.parse(open(fname).read())))
  return stmts

def column_args(column, pkey_fields):
  "parse sql column in a horrible way and return args for column constructor"
  parsed = column.parse()
  assert parsed.success # todo: does parse() log something on failure?
  dets = {'name': parsed.name, 'type': parsed.type}
  if parsed.pkey or parsed.name in pkey_fields:
    dets['primary_key'] = True
  if parsed.default:
    # todo: figure out escaping here
    dets['server_default'] = "sa.text('%s')" % parsed.default
  if parsed.unique:
    raise NotImplementedError('todo: sa_harness unique keys')
  if parsed.not_null:
    raise NotImplementedError('todo: sa_harness not null')
  return dets.pop('name'), dets.pop('type'), dets

TYPES = {
  'int': 'Integer',
  'jsonb': 'JSONB',
  'json': 'JSON',
  'text': 'Text',
  'uuid': 'UUID',
  'timestamp': 'DateTime',
}

def render_col(col_name, col_type, details, composite_pkey):
  if composite_pkey and col_name in composite_pkey:
    details['primary_key'] = True
  return f"sa.Column('{col_name}', sa.{TYPES[col_type]}{', ' + ', '.join('%s=%s' % pair for pair in details.items()) if details else ''})"

def transform(table_stmts, delim=',\n  '):
  "return some representation of sqlalchemy tables"
  table_strings = []
  for tablename, stmts in diffing.group_by_table(table_stmts).items():
    indexes = []
    table = None
    cols = collections.OrderedDict()
    composite_pkey = None
    for stmt in stmts:
      if isinstance(stmt, wrappers.CreateTable):
        pkey_fields = stmt.pkey_fields()
        for col in stmt.columns():
          name, type_, dets = column_args(col, pkey_fields)
          cols[name, type_] = dets
      elif isinstance(stmt, wrappers.CreateIndex):
        paren = stmt.decl()[-1]
        assert isinstance(paren, sqlparse.sql.Parenthesis)
        index_cols = ['%s_table.c.%s' % (stmt.table, ident) for ident, in wrappers.split_pun(paren)]
        indexes.append(f"{stmt.index_name}_index = sa.Index('{stmt.index_name}', {', '.join(index_cols)})")
      else:
        raise TypeError('unhandled statement type', type(stmt))
    if cols:
      col_strings = [
        render_col(col_name, col_type, details, composite_pkey)
        for (col_name, col_type), details in cols.items()
      ]
      table_strings.append(f"{tablename}_table = sa.Table('{tablename}', META{delim}{delim.join(col_strings)})")
    table_strings.extend(indexes)
  return table_strings

PARSER = argparse.ArgumentParser(description="generate sqlalchemy models from create table stmts")
PARSER.add_argument('glob')
PARSER.add_argument('--target', help="optional output path; otherwise stdout")

def main():
  args = PARSER.parse_args()
  stmts = read_glob_stmts(args.glob)
  value = '\n\n'.join([PREAMBLE] + transform(stmts))
  if args.target:
    open(args.target, 'w').write(value)
  else:
    print(value)

if __name__ == '__main__':
  main()



"wrappers.py -- wrappers around sqlparse to get expression info"

import sqlparse

class WrappedStatement:
  "base class"
  __slots__ = ('stmt',)

  def __init__(self, stmt):
    assert isinstance(stmt, sqlparse.sql.Statement)
    self.stmt = stmt

  def __eq__(self, other):
    "yes this is the way adults compare parse trees"
    return str(self.stmt).strip() == str(other.stmt).strip()

  def decl(self):
    "helper to get the function element, i.e. the table declaration"
    return next(expr for expr in self.stmt if isinstance(expr, sqlparse.sql.Function))

  @property
  def table(self):
    "return table name string"
    return self.decl().get_name()

def iswhitespace(token):
  "for splitting purposes, is this ignorable? includes whitespace & comments"
  # todo: do something else with comments, we probably want to decorate the parsed structure with special comments
  return token.is_whitespace or isinstance(token, sqlparse.sql.Comment)

def split_pun(tokens):
  """Takes a list of tokens and other stuff.
  Returns contiguous blocks of non-punctuation, i.e. list of lists.
  Recurses into IdentifierList which are inserted in insane places.
  Also strips whitespace tokens.
  """
  groups = [[]]
  for tok in tokens:
    if isinstance(tok, sqlparse.sql.Token) and tok.ttype and tok.ttype[0] == 'Punctuation':
      if groups[-1]: # i.e. if there's a non-empty group
        groups.append([])
    elif iswhitespace(tok):
      pass
    elif isinstance(tok, sqlparse.sql.IdentifierList):
      # this is insane
      inner = split_pun(tok)
      groups[-1].extend(inner[0])
      groups.extend(inner[1:])
      # note: intentionally not putting an empty list on the end. This is covered in test_diffing.py
    else:
      groups[-1].append(tok)
  if not groups[-1]:
    groups.pop()
  return groups

class ParsedColumn:
  # todo: py 3.7 dataclass
  __slots__ = ('success', 'name', 'type', 'default', 'unique', 'not_null', 'pkey')
  def __init__(self, success, name=None, type=None, default=None, unique=None, not_null=None, pkey=False):
    self.success = success
    self.name = name
    self.type = type
    self.default = default
    self.unique = unique
    self.not_null = not_null
    self.pkey = pkey

  def __eq__(self, other):
    # todo: py 3.7 dataclass provides this automatically
    if not isinstance(other, self.__class__):
      return False
    return tuple(map(lambda slot: getattr(self, slot), self.__slots__)) == \
      tuple(map(lambda slot: getattr(other, slot), other.__slots__))

  def __repr__(self):
    return f"<{self.__class__.__name__} {' '.join(map(lambda slot: f'{slot}={getattr(self, slot)}', self.__slots__))}>"

class Column:
  def __init__(self, tokens):
    self.tokens = tokens

  @property
  def name(self):
    assert isinstance(self.tokens[0], sqlparse.sql.Identifier)
    return self.tokens[0].value

  def __eq__(self, other):
    return [tok.value for tok in self.tokens] == [tok.value for tok in other.tokens]

  def render(self):
    return ' '.join(map(str, self.tokens))

  def parse(self):
    "return a ParsedColumn. This is rudimentary and will fail on hard cases"
    name, *tokens = self.tokens
    if not isinstance(name, sqlparse.sql.Identifier):
      return ParsedColumn(False)
    type_, *tokens = tokens
    if not isinstance(type_, sqlparse.sql.Function) and type_.ttype[-1] not in ('Builtin', 'Keyword'):
      return ParsedColumn(False)
    success = ParsedColumn(True, name.value, type_.value)
    if tokens and isinstance(tokens[0], sqlparse.sql.SquareBrackets):
      brackets, *tokens = tokens
      success.type += brackets.value
    while tokens:
      if tokens[0].ttype and tokens[0].ttype[-1] == 'Keyword':
        if tokens[0].normalized.lower() == 'default':
          _, val, *tokens = tokens
          success.default = val.value
        elif tokens[0].normalized.lower() == 'not null':
          _, *tokens = tokens
          success.not_null = True
        elif tokens[0].normalized.lower() == 'unique':
          _, *tokens = tokens
          success.unique = True
        elif tokens[0].normalized.lower() == 'primary':
          _, key, *tokens = tokens
          assert key.normalized.lower() == 'key'
          success.pkey = True
      else:
        return ParsedColumn(False)
    return success

def split_pun_paren(decl):
  """return things inside parenthesis, split by punctuation.
  decl is a 'function expression' i.e. a piece of a create table
  """
  paren = next(
    expr for expr in decl
    if isinstance(expr, sqlparse.sql.Parenthesis)
  )
  return split_pun(paren)


class CreateTable(WrappedStatement):
  def __init__(self, stmt):
    assert stmt.get_type() == 'CREATE'
    super().__init__(stmt)

  def groups(self):
    "helper for columns() and pkey_fields(). returns tokens inside the first paren, grouped by punctuation split."
    return split_pun_paren(self.decl())

  def columns(self):
    return [
      Column(group)
      for group in self.groups()
      # note: this guard blocks 'primary key (a,b)' kind of thing
      if isinstance(group[0], sqlparse.sql.Identifier)
    ]

  def pkey_fields(self):
    "return list of column names that are in primary key"
    for group in self.groups():
      if isinstance(group[0], sqlparse.sql.Token) and group[0].ttype and group[0].ttype[0] == 'Keyword':
        idents = [ident for ident, in split_pun_paren(group)]
        assert all(isinstance(ident, sqlparse.sql.Identifier) for ident in idents)
        return [str(ident) for ident in idents]
    else:
      return [col.name for col in self.columns() if col.parse().pkey]

  @property
  def unique(self):
    "unique key -- in this case, ('create', name) because we only allow one create stmt per table"
    return ('create', self.table)

  def tail(self):
    "non-space tokens after column parens"
    paren_index = next((i for i, expr in enumerate(self.stmt) if isinstance(expr, sqlparse.sql.Function)), None)
    if paren_index is None:
      return None
    tail = self.stmt[paren_index + 1:]
    return [expr for expr in tail if not iswhitespace(expr) and expr.value != ';']

class CreateIndex(WrappedStatement):
  def __init__(self, stmt):
    assert stmt.get_type() == 'CREATE'
    super().__init__(stmt)

  @property
  def index_name(self):
    index_index = next(i for i, tok in enumerate(self.stmt) if tok.ttype and tok.ttype[-1] == 'Keyword' and tok.value == 'index')
    on_index = next(i for i, tok in enumerate(self.stmt) if tok.ttype and tok.ttype[-1] == 'Keyword' and tok.value == 'on')
    block = self.stmt[index_index:on_index]
    ident, = next(tok for tok in block if isinstance(tok, sqlparse.sql.Identifier))
    return ident.value

  @property
  def unique(self):
    "unique key -- in this case, ('create', name) because we only allow one create stmt per table"
    return ('index', self.index_name)

def wrap(stmt):
  assert isinstance(stmt, sqlparse.sql.Statement)
  if stmt.get_type() == 'CREATE':
    keywords = [tok.value for tok in stmt if tok.ttype and tok.ttype[-1] == 'Keyword']
    if 'table' in keywords:
      return CreateTable(stmt)
    elif 'index' in keywords:
      return CreateIndex(stmt)
  elif stmt.get_type() == 'UNKNOWN':
    pass # cross fingers this means whitespace
  else:
    raise NotImplementedError(stmt.get_type(), stmt)



import pytest, sqlparse
from automig.lib import sa_harness, wrappers

def transform_string(string):
  "helper"
  return sa_harness.transform(map(wrappers.wrap, sqlparse.parse(string)), delim=', ')

def test_read_globs():
  "this one is just exercising"
  stmts = sa_harness.read_glob_stmts('test/schema/*.sql')

def test_transform_simple():
  sa_tables = transform_string('create table t1 (a int);')
  assert sa_tables == ["t1_table = sa.Table('t1', META, sa.Column('a', sa.Integer))"]
  # primary key
  sa_tables = transform_string('create table t1 (a int primary key, b int);')
  assert sa_tables == ["t1_table = sa.Table('t1', META, sa.Column('a', sa.Integer, primary_key=True), sa.Column('b', sa.Integer))"]
  # default
  sa_tables = transform_string('create table t1 (a timestamp default now());')
  assert sa_tables == ["t1_table = sa.Table('t1', META, sa.Column('a', sa.DateTime, server_default=sa.text('now()')))"]

def test_transform_index():
  sa_tables = transform_string('create index i1 on t1(a);')
  assert sa_tables == ["i1_index = sa.Index('i1', t1_table.c.a)"]

def test_transform_composite_key():
  # note: I think that still doesn't support this
  sa_tables = transform_string('create table t1 (a int, b int, primary key (a, b));')
  assert sa_tables == ["t1_table = sa.Table('t1', META, sa.Column('a', sa.Integer, primary_key=True), sa.Column('b', sa.Integer, primary_key=True))"]



#!/usr/bin/env python3
"entrypoint for automigrate command line"
# example:
# automig 218dd2c 'test/schema/*.sql' --initial
# automig 218dd2c...b5b40ce 'test/schema/*.sql'
# automig 218dd2c...HEAD 'test/schema/*.sql'

import argparse, git, os, yaml
from datetime import datetime
from .lib import ref_diff, githelp

def read_version():
  return open(os.path.join(os.path.dirname(__file__), 'VERSION')).read().strip()

def build_parser():
  version = read_version()
  parser = argparse.ArgumentParser(description=__doc__ + f" v{version}")
  parser.add_argument('ref', help="single git ref (i.e. sha) or sha1...sha2")
  parser.add_argument('glob', help="glob to grab paths, typicaly 'schema/*.sql' or something. use single-quotes so bash doesn't complete it")
  parser.add_argument('--initial', action='store_true', help="is this an initial commit (i.e. create metadata)")
  parser.add_argument('--update-meta', action='store_true', help="apply / update the preamble; this may only work with postgres")
  parser.add_argument('--opaque', action='store_true')
  return parser

# this creates the meta tables
PREAMBLE = """
-- meta_meta stores the meta version
create table automigrate_meta_meta (id serial primary key, meta_version int, applied timestamp default now());
insert into automigrate_meta_meta (meta_version) values (3);

-- meta stores the schema version
create table automigrate_meta (id serial primary key, sha text, applied timestamp default now(), fromsha text, opaque bool, automig_version text);
"""

POSTGRES_PREAMBLE = """
-- meta_meta stores the meta version
create table if not exists automigrate_meta_meta (id serial primary key, meta_version int);
alter table automigrate_meta_meta add column if not exists applied timestamp default now();
insert into automigrate_meta_meta (meta_version) values (3);

-- meta stores the schema version
create table if not exists automigrate_meta (id serial primary key, sha text, applied timestamp default now());
alter table automigrate_meta add column if not exists fromsha text;
alter table automigrate_meta add column if not exists opaque bool;
alter table automigrate_meta add column if not exists automig_version text;
"""

def main():
  args = build_parser().parse_args()
  rev_tuple = githelp.parse_rev_to_tuple(args.ref)
  print(f'-- changeset created from {args} at {datetime.now()}')
  shas = []
  manual_mig = {}
  if os.path.exists('.manualmig.yml'):
    # todo: read this from repo root, not cwd
    manual_mig = yaml.safe_load(open('.manualmig.yml'))['overrides']
  if args.initial:
    if args.opaque:
      raise ValueError("don't pass --opaque with --initial")
    shas.append(None)
    print(PREAMBLE)
    commit = git.Repo(search_parent_directories=True).commit(rev_tuple[0])
    shas.append(commit.hexsha)
    tree = commit.tree
    assert len(rev_tuple) == 1, "can't pass a sha range for --initial"
    for contents in githelp.get_streams(tree, args.glob):
      print(contents.decode())
  else:
    assert len(rev_tuple) == 2, "must pass a sha range or set --initial"
    # todo: look up rev_tuple[0] so this isn't a short sha or HEAD~1 or something
    shas.append(rev_tuple[0])
    changes = ref_diff.ref_range_diff(git.Repo(search_parent_directories=True), *rev_tuple, args.glob, opaque=args.opaque)
    errors = ref_diff.extract_errors(changes)
    if errors:
      # todo: this is wrong. manual_mig should override whether or not there are errors
      remaining = ref_diff.try_repair_errors(errors, manual_mig, changes)
      if remaining:
        raise ValueError('errors not overridden in .manualmig.yml', remaining)
    if args.update_meta:
      print(POSTGRES_PREAMBLE)
    for sha, tables in changes.items():
      if not tables:
        continue
      shas.append(sha)
      for table, stmts in tables.items():
        print(f'-- changes for {sha[:7]}.{table}')
        for stmt in stmts:
          print(stmt)
  for sha1, sha2 in zip(shas[:-1], shas[1:]):
    # todo: hard w/out depending on a sql driver but do some kind of escaping here
    sha1 = f"'{sha1}'" if sha1 is not None else 'NULL'
    print(f"insert into automigrate_meta (fromsha, sha, automig_version, opaque) values ({sha1}, '{sha2}', '{read_version()}', {['false', 'true'][args.opaque]});")

if __name__ == '__main__':
  main()



import git, glob, pytest, sqlparse, os
from automig.lib import githelp, ref_diff, diffing

SHAS = {
  'create-t1': '218dd2c',
  'add-t1-col': '9dcbd4e81e9a0dd7629ed7ae82a86891a88f76f3',
  'add-t2-t1a': 'b5b40ce718ea7241fee8d0a3826f244d21bf413c',
  'unsup-alter-col': '0756d87',
}
GLOB ='test/schema/*.sql'

def test_get_paths():
  repo = git.Repo()
  tree = repo.commit(SHAS['create-t1']).tree
  assert githelp.get_paths(tree, os.path.join(repo.working_dir, GLOB)) == ['test/schema/sql.sql']

@pytest.mark.skip
def test_create():
  diff = ref_diff.ref_range_diff(git.Repo(), SHAS['create-t1'], SHAS['add-t1-col'], GLOB)
  raise NotImplementedError

def test_addcol():
  diff = ref_diff.ref_range_diff(git.Repo(), SHAS['create-t1'], SHAS['add-t1-col'], GLOB)
  assert diff == {
    SHAS['add-t1-col']: {'t1': ['alter table t1 add column b int;']},
  }

def test_add_multi_commit():
  diff = ref_diff.ref_range_diff(git.Repo(), SHAS['create-t1'], SHAS['add-t2-t1a'], GLOB)
  assert diff == {
    SHAS['add-t1-col']: {
      't1': ['alter table t1 add column b int;'],
    },
    SHAS['add-t2-t1a']: {
      't1': ['create index t1a on t1 (a);'],
      't2': ['create table t2 (a int primary key);'],
    },
  }

def test_add_multi_commit_opaque():
  diff = ref_diff.ref_range_diff(git.Repo(), SHAS['create-t1'], SHAS['add-t2-t1a'], GLOB, opaque=True)
  assert diff == {SHAS['add-t2-t1a']: {
    't1': ['alter table t1 add column b int;', 'create index t1a on t1 (a);'],
    't2': ['create table t2 (a int primary key);'],
  }}

MOD_COLUMN = [
  'create table t1 (a int primary key, b int);',
  'create table t1 (a int primary key, b int unique);',
]

def test_error_bubbling():
  sha_table_stmts = {'sha': diffing.diff(*map(sqlparse.parse, MOD_COLUMN))}
  errors = ref_diff.extract_errors(sha_table_stmts)
  manual = {'sha': {'t1': ['hello']}}
  remaining = ref_diff.try_repair_errors(errors, manual, sha_table_stmts)
  assert not remaining
  assert sha_table_stmts['sha']['t1'] == ['hello']



#!/usr/bin/env python3

import os
from distutils.core import setup
from setuptools import find_packages

setup(
  name='automig',
  version=open(os.path.join(os.path.dirname(__file__), 'automig', 'VERSION')).read().strip(),
  description="Declarative, automatic db migrations using SQL & git as the source of truth",
  author="Abe Winter",
  url="https://github.com/abe-winter/automigrate",
  packages=find_packages(include=['automig', 'automig.*']),
  entry_points = {
    'console_scripts': ['automig=automig.__main__:main'],
  },
  keywords=['sql', 'migration', 'git', 'diff'],
  install_requires=[
    'sqlparse==0.3.0',
    'gitpython==2.1.11',
    'pyyaml==5.1',
  ],
  python_requires='>=3.6', # for format strings
  long_description=open(os.path.join(os.path.dirname(__file__), 'README.md')).read(),
  long_description_content_type='text/markdown',
  include_package_data=True,
)



"githelp.py -- helpers for doing things on git"

import git, fnmatch, os, collections

def parse_rev_to_tuple(rev_string):
  if '...' in rev_string:
    left, _, right = rev_string.partition('...')
    return left, right
  elif '..' in rev_string:
    raise ValueError("we don't support two-dot strings -- use a three-dot")
  else:
    return rev_string,

def get_paths(tree, pattern, root=()):
  """given tree (a git tree at a desired ref) return list of paths matching pattern
  note: this matches paths that existed at the time of ref, including files that don't exist now and excluding ones that exist on disk but were missing at ref
  """
  found = []
  for item in tree:
    if isinstance(item, git.Blob):
      if fnmatch.fnmatch(item.abspath, pattern):
        found.append(item.path)
    elif isinstance(item, git.Tree):
      found.extend(get_paths(item, pattern, root + (item.name,)))
    else:
      raise NotImplementedError("unexpected type in tree", item)
  return found

def get_streams(tree, pattern):
  """given ref (git sha or reference as string) and glob (glob pattern as string)
  return list of file contents
  todo: support repos you're not inside of
  """
  return [tree[path].data_stream.read() for path in get_paths(tree, pattern)]
